/*
 * Copyright (c) 2012 Alexandre Tuleu
 *
 * This file is part of xeno-cpg.
 *
 *   xeno-cpg is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   xeno-cpg is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with PROJECT.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \file CpgNetworkRTWrapper.cpp
 *
 *  \date Jul 1, 2011
 *  \author Alexandre Tuleu
 */

#include <iostream>
#include <cmath>
#include <stdexcept>

#include <cpg-network/cpg-network.h>

#include <native/mutex.h>

#include "CpgNetworkRTWrapper.h"
#include "XenoCpgConfig.h"
#include "RTIO.h"




RT_MUTEX CpgMutex;
const char * CpgNetworkRTWrapper::s_mutexName("Xeno CPG Network muetx");

bool CpgNetworkRTWrapper::operator()() const{
	return (d_network != NULL )&& d_valid;
}


CpgNetworkRTWrapper::CpgNetworkRTWrapper()
	: d_valid(false)
	, d_network(NULL)
	, d_iterationPerStep(1)
	, d_ts(0.005)
	, d_initTime(0.0)
	, d_time(0.0)
	, d_paused(true){
}

CpgNetworkRTWrapper::~CpgNetworkRTWrapper(){
	DeInit();
}


void CpgNetworkRTWrapper::Compile(){
	if(d_network == NULL){
		throw std::runtime_error("The nework is not initialized. Please call Init().");
	}
	cpg_object_compile(CPG_OBJECT(d_network),NULL,NULL);
}

void CpgNetworkRTWrapper::Start(){
	d_paused = false;
}

void CpgNetworkRTWrapper::Pause(){
	d_paused = true;
}

int CpgNetworkRTWrapper::Step(){
	if(!(*this)()){
		std::cerr << "Bad cpg network status ! " << std::endl;
		return 1;
	}

	if(!d_paused){
		if(d_time >= d_initTime){
			rt_mutex_acquire(&CpgMutex,TM_INFINITE);
			UploadValuesToCPG();
			for(unsigned int i = 0; i < d_iterationPerStep; ++i){
				cpg_network_step(d_network,d_ts);
			}
			rt_mutex_release(&CpgMutex);
		} else {
			d_time += d_ts * d_iterationPerStep;
		}
	}


	rt_mutex_acquire(&CpgMutex,TM_INFINITE);
	for(OutputPropertyMapper::iterator p = d_outputs.begin();
	    p != d_outputs.end();
	    ++p){
		double val = cpg_property_get_value(p->first);
		p->second->SetValue(val);
	}
	rt_mutex_release(&CpgMutex);

	return 0;
}


void CpgNetworkRTWrapper::Init(){
	if(d_network != NULL){
		throw std::runtime_error("You should not Init twice");
	}
	g_type_init();

	GError *error = NULL;

	CpgSection & cpg = XenoCpgConfig::Instance().Cpg();

	d_ts = cpg.Timestep();
	d_iterationPerStep = cpg.IterationPerTimestep();

	d_network = cpg_network_new_from_path(cpg.NetworkFile().c_str(),
	                                      &error);

	if(!d_network){
		std::cerr << "Unable to load network file '"
		          << cpg.NetworkFile() << "' : "
		          << error->message << std::endl;
		g_error_free(error);
		d_valid = false;
		return;
	}

	InitMappingToServo();
	InitNetworkOverride();
	InitGlobals();

	d_valid = true;
}

void CpgNetworkRTWrapper::DeInit(){
	if(d_network){
		g_object_unref(d_network);
	}
}

CpgProperty *
CpgNetworkRTWrapper::FindPropertyFromSerialized(const std::string & serialized,
                                          std::string * canonicalName){
	std::string::size_type pos = serialized.find("::");

	CpgObject *obj;
	std::string propName,objName;

	if(pos !=std::string::npos) {
		objName = serialized.substr(0,pos);
		obj = cpg_group_find_object((CpgGroup*)(d_network),
				objName.c_str());
		propName = serialized.substr(pos + 2);
	} else {
		objName = "globals";
		obj = (CpgObject*)(d_network);//cpg_object_get_globals(d_network);
		propName = serialized;
	}

	if(canonicalName != 0){
		*canonicalName = objName;
		canonicalName->append(".").append(propName);
	}
	if(!obj) {
		std::cerr << "Unable to find object '" << objName
		          << "' in network !" << std::endl;
		return NULL;
	}

	CpgProperty *prop = cpg_object_get_property(obj,propName.c_str());

	if(!prop){
		std::cerr << "Unable to find property '" << propName << "' in object '"
		          << objName << "'." << std::endl;
		return NULL;
	}
	return prop;
}

void CpgNetworkRTWrapper::InitMappingToServo(){
	typedef options::Section::ListOfOptions ListOfOptions;
	const ListOfOptions & mapped = XenoCpgConfig::Instance().Cpg().OutMapping()->Options();

	RTIO & io = RTIO::Instance();

	std::string canonName;
	for(ListOfOptions::const_iterator prop = mapped.begin();
	    prop != mapped.end();
	    ++prop){
		CpgOutput::Ptr out = io.Output(prop->second->Name());

		if(!out){
			std::cerr << "No output '" << prop->second->Name()
			          << "' registered." << std::endl;
			continue;
		}

		CpgProperty *property(FindPropertyFromSerialized(prop->second->ValueAsString(),&canonName));
		if(!property){
			continue;
		}
		d_outputs.insert(std::make_pair(property,out));
		std::cout << "<<< Mapping " << canonName << " to actuator '"
		          << prop->second->Name() << "'."
		          << std::endl;
	}
}


void CpgNetworkRTWrapper::InitNetworkOverride(){
	typedef options::Section::ListOfOptions ListOfOverrides;

	ListOfOverrides & overrides = XenoCpgConfig::Instance().Cpg().CpgOverride()->Options();
	std::string canoName;
	for(ListOfOverrides::const_iterator ov = overrides.begin();
	    ov != overrides.end();
	    ++ov){

		if(ov->first.empty()){
			continue;
		}
		CpgProperty *prop(FindPropertyFromSerialized(ov->first,&canoName));
		if(prop){
			const char * expression = ov->second->ValueAsString().c_str();
			cpg_property_set_expression(prop,
			                            cpg_expression_new(expression));
			std::cout << "<<< overriding '" << canoName << "' with : " << ov->second->ValueAsString()
			          << "." << std::endl;
		}
	}
}



void CpgNetworkRTWrapper::InitGlobals(){
	if((*this)()){
		return;
	}

	CpgObject * globals = CPG_OBJECT(d_network);

	GSList * propListPtr = cpg_object_get_properties(globals);

	while(propListPtr != NULL){
		CpgProperty * p = CPG_PROPERTY(propListPtr->data);

		if(cpg_property_get_flags(p) & CPG_PROPERTY_FLAG_IN){
			std::string propName(cpg_property_get_name(p));
			d_globals.push_back(CpgPropertyIO::Ptr(new CpgPropertyIO(p,
			                                                         propName)));
		}
		propListPtr = propListPtr->next;
	}
}


void CpgNetworkRTWrapper::UploadValuesToCPG() {
	for(InputPropertyMapper::iterator input = d_inputs.begin();
		input != d_inputs.end();
		input++){
		cpg_property_set_value(input->first, input->second->Value());
	}
}


const CpgNetworkRTWrapper::CpgPropertyIOList & CpgNetworkRTWrapper::GlobalInputProperty(){
	return d_globals;
}

CpgPropertyIO::CpgPropertyIO(CpgProperty* property , const std::string& name)
	: d_property(property)
	, d_name(name){
}

CpgPropertyIO::~CpgPropertyIO(){
}

void CpgPropertyIO::Set(double value){
	rt_mutex_acquire(&CpgMutex,TM_INFINITE);
	cpg_property_set_value(d_property,value);
	rt_mutex_release(&CpgMutex);
}

double CpgPropertyIO::Read() const{
	rt_mutex_acquire(&CpgMutex,TM_INFINITE);
	double val = cpg_property_get_value(d_property);
	rt_mutex_release(&CpgMutex);
	return val;
}

const std::string& CpgPropertyIO::Name() const{
	return d_name;
}

bool CpgNetworkRTWrapper::Valid(){
	return (*this)();
}


void CpgNetworkRTWrapper::InitRtAccess(){

	int res;
	if((res = rt_mutex_create(&CpgMutex,s_mutexName)) != 0){

		switch(res) {
		case -ENOMEM :
			throw std::runtime_error("system fails to get enough dynamic memory from the global real-time heap in order to register the mutex");
		case -EEXIST :
			throw std::runtime_error("the name is already in use by some registered object");
		case -EPERM :
			throw std::runtime_error("this service was called from an asynchronous context");
		default :
			std::ostringstream os;
			os <<"undocummented error code : " << res;
			throw std::runtime_error(os.str());
		}

	}
}

void CpgNetworkRTWrapper::WaitForRTAccessInitialized(){

	RT_MUTEX foo;
	int res = rt_mutex_bind(&foo,s_mutexName,TM_INFINITE);
	if(res != 0){
		switch(res){
		case -EFAULT :
			throw std::runtime_error("mutex or name is referencing invalid memory");
		case -EINTR :
			throw std::runtime_error("rt_task_unblock() has been called for the waiting task before the retrieval has completed");
		case -EWOULDBLOCK :
			throw std::runtime_error("timeout is equal to TM_NONBLOCK and the searched object is not registered on entry");
		case -ETIMEDOUT :
			throw std::runtime_error("the object cannot be retrieved within the specified amount of time");
		case -EPERM :
			throw std::runtime_error("this service should block, but was called from a context which cannot sleep (e.g. interrupt, non-realtime context)");
		default :
			std::ostringstream os;
			os <<"undocummented error code : " << res;
			throw std::runtime_error(os.str());
		}
	}
	//unbind mutex !!!
	rt_mutex_unbind(&foo);

}

CpgIOBase::CpgIOBase()
	: d_value(0){
}

CpgIOBase::~CpgIOBase(){}

double CpgIOBase::Value() const {
	return d_value;
}

void CpgIOBase::SetValue(double value){
	d_value = value;
}

CpgOutput::CpgOutput(){}

CpgOutput::~CpgOutput(){}

CpgInput::CpgInput(){}

CpgInput::~CpgInput(){}
