/*
 * Copyright (c) 2012 Alexandre Tuleu
 *
 * This file is part of xeno-cpg.
 *
 *   xeno-cpg is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   xeno-cpg is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with PROJECT.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \file XenoCpg.cpp
 *
 *  \date Jun 18, 2012
 *  \author Alexandre Tuleu
 */

#include <stdexcept>

#include "XenoCpg.h"

#include "XenoCpgConfig.h"
#include "RTIO.h"
#include "XenoCpgPrompt.h"
#include <boost/program_options.hpp>


const char * XenoCpg::s_initTaskName("xeno-cpg initialization task");
const char * XenoCpg::s_mainTaskName("xeno-cpg main task");
const char * XenoCpg::s_promptTaskName("xeno-cpg prompt task");


void XenoCpg::Initializer::Init(){}

XenoCpg::XenoCpg():d_prompted(false){}

XenoCpg::~XenoCpg(){}

void XenoCpg::Init(const std::string& name , int& argc , char** argv){
	XenoCpgConfig::Instance().SetBaseFileName(name);

	typedef std::vector<std::string> ListOfFiles;
	using namespace boost::program_options;
	options_description desc("Allowed options");
	desc.add_options()
		("help,h", "produce help message")
		("config,C", value< ListOfFiles >(), "config file to append");

	variables_map vm;
	store(parse_command_line(argc,argv,desc),vm);
	notify(vm);

	if(vm.count("help")){
		std::cout << desc << std::endl;
		exit(0);
	}

	XenoCpgConfig::Instance().LoadStdFiles();

	if(vm.count("config")){
		const ListOfFiles & files(vm["config"].as<ListOfFiles>());

		for(ListOfFiles::const_iterator f =  files.begin();
		    f != files.end();
		    ++f){
			XenoCpgConfig::Instance().LoadFile(*f);
		}
	}


	rt_task_create(&d_main_task,s_mainTaskName,0,90,0);

	rt_task_create(&d_prompt_task,s_promptTaskName,0,70,T_JOINABLE);
}

void XenoCpg::StartMainLoop(){
	rt_task_start(&d_main_task,&XenoCpg::MainTask,NULL);
}

void XenoCpg::StartPrompt(){
	signal(SIGTERM,&XenoCpg::CatchSignal);
	signal(SIGINT,&XenoCpg::CatchSignal);


	rt_task_start(&d_prompt_task,&XenoCpg::PromptTask,NULL);
	d_prompted = true;
}

bool XenoCpg::ThreadPromptStarted(){
	return d_prompted = true;
}

void XenoCpg::CatchSignal(int sig){
	XenoCpg & c = XenoCpg::Instance();
	if(c.d_prompted && rt_task_inquire(&(c.d_prompt_task),NULL) == 0 ){
		rt_task_delete(&(c.d_prompt_task));
	}
}

void XenoCpg::WaitForCompletion(){
	if(ThreadPromptStarted()){
		rt_task_join(&d_prompt_task);
	} else {
		pause();
	}
	rt_task_delete(&d_main_task);

	if(rt_task_inquire(&d_prompt_task,NULL) == 0 ){
		rt_task_delete(&d_prompt_task);
	}
}

void XenoCpg::AppendInitializer(Initializer::Ptr init){
	d_initializers.push_back(init);
}

void XenoCpg::MainTask(void* controller){
	XenoCpg & x = XenoCpg::Instance();
	for(ListOfInitializer::iterator i = x.d_initializers.begin();
	    i != x.d_initializers.end();
	    ++i){
		(*i)->Init();
	}

	x.d_network.Init();

	x.d_network.Compile();

	if(!x.d_network()){
		throw std::runtime_error("Cpg Network file seems unvalid.");
	}

	CpgNetworkRTWrapper::InitRtAccess();

	CpgNetworkRTWrapper & network = XenoCpg::Instance().d_network;

	RTIO & io = RTIO::Instance();

	double ts = XenoCpgConfig::Instance().Cpg().Timestep();

	unsigned int nbIteration,iterForFeedback(XenoCpgConfig::Instance().Main().CliFeedbackIteration());

	int res = rt_task_set_periodic(NULL,TM_NOW,ts * 1e9);

	switch(res){
	case 0 :
		break;
	case -EINVAL :
		throw std::runtime_error("task is not a task descriptor, or period is different from TM_INFINITE but shorter than the scheduling latency value for the target system, as available from /proc/xenomai/latency");
	case -EIDRM :
		throw std::runtime_error("task is a deleted task descriptor");
	case -ETIMEDOUT :
		throw std::runtime_error("idate is different from TM_INFINITE and represents a date in the past");
	case -EWOULDBLOCK :
		throw std::runtime_error("the system timer is not active");
	case -EPERM :
		throw std::runtime_error("task is NULL but not called from a task context");
	default :
		std::ostringstream os;
		os <<"undocummented error code : " << res;
		throw std::runtime_error(os.str());
	}
	unsigned long overruns;
	while(true){
		res = rt_task_wait_period(&overruns);
		switch(res){
		case 0 :
		case -ETIMEDOUT :
			break;
		case -EWOULDBLOCK :
			throw std::runtime_error("rt_task_set_periodic() has not previously been called for the calling task");
		case -EINTR :
			throw std::runtime_error("rt_task_unblock() has been called for the waiting task before the next periodic release point has been reached. In this case, the overrun counter is reset too");
		case -EPERM :
			throw std::runtime_error("this service was called from a context which cannot sleep (e.g. interrupt, non-realtime or scheduler locked)");
		default :
			std::ostringstream os;
			os <<"undocummented error code : " << res;
			throw std::runtime_error(os.str());
		}

		if(overruns > 0){
			std::cerr << "Got " << overruns << " overrun"
					<< (overruns > 1 ? "s" : "")
					<< "." << std::endl;
		}

		for(unsigned long i = 0 ; i < overruns +1;++i){
			network.Step();
		}

		io.Sync();
		nbIteration += overruns +1;

		if(iterForFeedback > 0){
			if(nbIteration > iterForFeedback){
				io.Print(std::cout);
				nbIteration = 0;
			}
		}

	}
}

void XenoCpg::PromptTask(void* ){

	CpgNetworkRTWrapper::WaitForRTAccessInitialized();

	CpgNetworkRTWrapper & networkWrapper = XenoCpg::Instance().d_network;

	cli::Interface::Initialize(new XenoCpgPrompt(networkWrapper));

	while(true){
		if(cli::Interface::Instance().Iteration()){
			break;
		}
		usleep(10000);//10ms
	}
}


