/* 
 * Copyright (c) 2012 Alexandre Tuleu
 * 
 * This file is part of xeno-cpg.
 * 
 *   xeno-cpg is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 * 
 *   xeno-cpg is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 * 
 *   You should have received a copy of the GNU General Public License
 *   along with PROJECT.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \file XenoCpgPrompt.cpp
 *
 *  \date May 2, 2012
 *  \author Alexandre Tuleu
 */

#include "XenoCpgPrompt.h"

#include <fstream>

XenoCpgPrompt::XenoCpgPrompt(CpgNetworkRTWrapper & controller)
	: d_controller(controller){
}

XenoCpgPrompt::~XenoCpgPrompt(){
}


void XenoCpgPrompt::Init(){
	const CpgNetworkRTWrapper::CpgPropertyIOList & globals = d_controller.GlobalInputProperty();


	for(CpgNetworkRTWrapper::CpgPropertyIOList::const_iterator p = globals.begin();
	    p != globals.end();
	    ++p){
		PropertyVariable::Ptr pv(new PropertyVariable(*p));
		this->AddVariable(pv);
		d_cpgProperties.insert(std::make_pair(pv->Name(),pv));
	}

	this->AddCommand(cli::Closure1<XenoCpgPrompt,std::string>::Create("save",*this,&XenoCpgPrompt::Save));
	this->AddCommand(cli::Closure2<XenoCpgPrompt,std::string,double>::Create("select",*this,&XenoCpgPrompt::Select));
	this->AddCommand(cli::Closure0<XenoCpgPrompt>::Create("unselect",*this,&XenoCpgPrompt::Unselect));
	this->AddCommand(cli::Closure0<XenoCpgPrompt>::Create("start",*this,&XenoCpgPrompt::Start));
	this->AddCommand(cli::Closure0<XenoCpgPrompt>::Create("pause",*this,&XenoCpgPrompt::Pause));

}

XenoCpgPrompt::PropertyVariable::PropertyVariable(CpgPropertyIO::Ptr var)
	: VariableBase(var->Name())
	, d_value(var->Read())
	, d_io(var){
	SetStringFromValue();
}

XenoCpgPrompt::PropertyVariable::~PropertyVariable(){
}

void XenoCpgPrompt::PropertyVariable::PreGetValueHook() const{
	const_cast<PropertyVariable*>(this)->d_value = d_io->Read();
	const_cast<PropertyVariable*>(this)->SetStringFromValue();
}

void XenoCpgPrompt::PropertyVariable::PostSetValueHook(){
	d_io->Set(d_value);
	SetStringFromValue();
}

const std::string& XenoCpgPrompt::PropertyVariable::toString() const{
	this->PreGetValueHook();
	return d_stringValue;
}

bool XenoCpgPrompt::PropertyVariable::SetFromString(const std::string& val){
	double actualValue;
	std::istringstream iss(val);
	iss >> actualValue;
	if(iss.fail()){
		return false;
	}
	d_value = actualValue;
	this->PostSetValueHook();
	return true;
}

void XenoCpgPrompt::PropertyVariable::Increment(double value){
	d_value += value;
	PostSetValueHook();
}

void XenoCpgPrompt::PropertyVariable::Decrement(double value){
	d_value -= value;
	PostSetValueHook();
}



void XenoCpgPrompt::Save(const std::string & filename){

	std::ofstream file(filename.c_str());

	if(!file) {
		throw std::runtime_error("Unable to open file '" + filename + "'");
	}

	file << "[cpg.override]" << std::endl<< std::endl;


	for(VariableByName::const_iterator pv =Variables().begin();
		pv != Variables().end();
		++pv){
		file << pv->second->Name() << " = " << pv->second->toString();
	}
}


XenoCpgPrompt::IncrementSelected::IncrementSelected(
        PropertyVariable::Ptr pv , double increment , bool decrementInstead){
}

XenoCpgPrompt::IncrementSelected::~IncrementSelected(){
}

int XenoCpgPrompt::IncrementSelected::Call(int key){
	if(!d_pv || !d_pv.get()){
		return 1;
	}

	if(d_decrement){
		d_pv->Decrement(d_increment);
	} else {
		d_pv->Increment(d_increment);
	}
	return 0;
}

XenoCpgPrompt::PropertyVariable::Ptr
XenoCpgPrompt::PropertyVariableFromName(const std::string& name){
	PropertyVariableByName::const_iterator fi = d_cpgProperties.find(name);
	if ( fi == d_cpgProperties.end()){
		return PropertyVariable::Ptr();
	}
	return fi->second;
}

void XenoCpgPrompt::PropertyVariable::SetStringFromValue(){
	std::ostringstream os;
	os << d_value;
	d_stringValue = os.str();
}

void XenoCpgPrompt::Select(const std::string & propName , const double & increment){
	d_selected = PropertyVariableFromName(propName);
	if(!d_selected || !d_selected.get()){
		throw std::runtime_error("could not find globals '" + propName + "' for selection");
	}

	cli::KeyCallback::Ptr incr(new IncrementSelected(d_selected,increment,false));
	cli::KeyCallback::Ptr decr(new IncrementSelected(d_selected,increment,false));
	this->AddBinding(IncrementKey,incr);
	this->AddBinding(DecrementKey,decr);
	this->SetPrompt(d_selected->Name() + " >>> ");
}

void XenoCpgPrompt::Unselect(){
	d_selected = PropertyVariable::Ptr();
	this->Unbind(IncrementKey);
	this->Unbind(DecrementKey);
	this->SetPrompt(">>> ");
}

void XenoCpgPrompt::Start(){
	d_controller.Start();
}

void XenoCpgPrompt::Pause(){
	d_controller.Pause();
}

